from __future__ import annotations

import functools
from typing import Callable, Literal, Optional, cast

import panel as pn
import param
from panel.reactive import ReactiveHTML

from ragna._compat import anext

from . import styles as ui

# TODO : move all the CSS rules in a dedicated file

message_stylesheets = [
    """ 
            :host .right, :host .center {
                    width:100% !important;
            }
    """,
    """ 
            :host .left {
                height: unset !important;
                min-height: unset !important;
            }
    """,
    """
            :host div.bk-panel-models-layout-Column:not(.left) { 
                    width:100% !important;
            }
    """,
    """
            :host .message {
                width: calc(100% - 15px);
                box-shadow: unset;
                font-size: unset;
                background-color: unset;
            }
    """,
    """
            :host .avatar {
                margin-top:0px;
                box-shadow: unset;
            }
    """,
]


class CopyToClipboardButton(ReactiveHTML):
    title = param.String(default=None, doc="The title of the button ")
    value = param.String(default=None, doc="The text to copy to the clipboard.")

    _template = """
        <div type="button" 
                id="copy-button"
                onclick="${script('copy_to_clipboard')}"
                class="container"
                style="cursor: pointer;"
        >
            <svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-clipboard" width="16" height="16" 
                    viewBox="0 0 24 24" stroke-width="2" stroke="gray" fill="none" stroke-linecap="round" stroke-linejoin="round">
                <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
                <path d="M9 5h-2a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h10a2 2 0 0 0 2 -2v-12a2 2 0 0 0 -2 -2h-2" />
                <path d="M9 3m0 2a2 2 0 0 1 2 -2h2a2 2 0 0 1 2 2v0a2 2 0 0 1 -2 2h-2a2 2 0 0 1 -2 -2z" />
            </svg>
            <span>${title}</span>
        </div>            
        """

    _scripts = {
        "copy_to_clipboard": """navigator.clipboard.writeText(`${data.value}`);"""
    }

    _stylesheets = [
        """ div.container { 
                            
                            display:flex;
                            flex-direction: row;
                            margin: 7px 10px;
                    
                            color:gray;
                        } 
                    
                        svg {
                            margin-right:5px;
                        }

                    """
    ]


class RagnaChatMessage(pn.chat.ChatMessage):
    _stylesheets = [
        *pn.chat.ChatMessage._stylesheets,
        *message_stylesheets,
    ]

    role: str = param.Selector(objects=["system", "user", "assistant"])
    sources = param.List(allow_None=True)
    on_click_source_info_callback = param.Callable(allow_None=True)

    def __init__(
        self,
        content: str,
        *,
        role: Literal["system", "user", "assistant"],
        user: str,
        sources: Optional[list[dict]] = None,
        on_click_source_info_callback: Optional[Callable] = None,
        timestamp=None,
        show_timestamp=True,
    ):
        css_class = f"message-content-{self.role}"
        self.content_pane = pn.pane.Markdown(
            content,
            css_classes=["message-content", css_class],
            stylesheets=ui.stylesheets(
                (
                    "table",
                    {"margin-top": "10px", "margin-bottom": "10px"},
                )
            ),
        )

        if role == "assistant":
            assert sources is not None
            css_class = "message-content-assistant-with-buttons"
            object = pn.Column(
                self.content_pane,
                self._copy_and_source_view_buttons(),
                css_classes=[css_class],
            )
        else:
            object = self.content_pane

        object.stylesheets.extend(
            ui.stylesheets(
                (
                    f":host(.{css_class})",
                    {"background-color": "rgb(243, 243, 243) !important"}
                    if role == "user"
                    else {
                        "background-color": "none",
                        "border": "rgb(234, 234, 234)",
                        "border-style": "solid",
                        "border-width": "1.2px",
                        "border-radius": "5px",
                    },
                )
            ),
        )

        super().__init__(
            object=object,
            role=role,
            user=user,
            sources=sources,
            on_click_source_info_callback=on_click_source_info_callback,
            timestamp=timestamp,
            show_timestamp=show_timestamp,
            show_reaction_icons=False,
            show_user=False,
            show_copy_icon=False,
            css_classes=[f"message-{role}"],
            avatar_lookup=functools.partial(self._ragna_avatar_lookup, role=role),
        )

    def _copy_and_source_view_buttons(self) -> pn.Row:
        return pn.Row(
            CopyToClipboardButton(
                value=self.content_pane.object,
                title="Copy",
                stylesheets=[
                    ui.CHAT_INTERFACE_CUSTOM_BUTTON,
                ],
            ),
            pn.widgets.Button(
                name="Source Info",
                icon="info-circle",
                stylesheets=[
                    ui.CHAT_INTERFACE_CUSTOM_BUTTON,
                ],
                on_click=lambda event: self.on_click_source_info_callback(
                    event, self.sources
                ),
            ),
        )

    # This cannot be a bound method, because it creates a reference cycle when trying
    # to access the repr of the message. See
    # https://github.com/Quansight/ragna/issues/359 for details.
    @staticmethod
    def _ragna_avatar_lookup(user: str, *, role: str) -> str:
        if role == "system":
            return "imgs/ragna_logo.svg"
        elif role == "user":
            return "👤"

        try:
            organization, model = user.split("/")
        except ValueError:
            organization = ""
            model = user

        if organization == "Ragna":
            return "imgs/ragna_logo.svg"
        elif organization == "OpenAI":
            if model.startswith("gpt-3"):
                return "https://upload.wikimedia.org/wikipedia/commons/thumb/0/04/ChatGPT_logo.svg/1024px-ChatGPT_logo.svg.png?20230318122128"
            elif model.startswith("gpt-4"):
                return "https://upload.wikimedia.org/wikipedia/commons/a/a4/GPT-4.png"
        elif organization == "Anthropic":
            return "https://upload.wikimedia.org/wikipedia/commons/1/14/Anthropic.png"

        return model[0].upper()


class RagnaChatInterface(pn.chat.ChatInterface):
    get_user_from_role = param.Callable(allow_None=True)

    @param.depends("placeholder_text", watch=True, on_init=True)
    def _update_placeholder(self):
        self._placeholder = RagnaChatMessage(
            ui.message_loading_indicator,
            role="system",
            user=self.get_user_from_role("system"),
            show_timestamp=False,
        )

    def _build_message(self, *args, **kwargs) -> RagnaChatMessage | None:
        message = super()._build_message(*args, **kwargs)
        if message is None:
            return None

        # We only ever hit this function for user inputs, since we control the
        # generation of the system and assistant messages manually. Thus, we can
        # unconditionally create a user message here.
        return RagnaChatMessage(message.object, role="user", user=self.user)


class CentralView(pn.viewable.Viewer):
    current_chat = param.ClassSelector(class_=dict, default=None)

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)

        # FIXME: make this dynamic from the login
        self.user = ""
        self.api_wrapper = api_wrapper
        self.chat_info_button = pn.widgets.Button(
            # The name will be filled at runtime in self.header
            name="",
            on_click=self.on_click_chat_info_wrapper,
            button_style="outline",
            icon="info-circle",
            stylesheets=[":host { margin-top:10px; }"],
        )
        self.on_click_chat_info = None

    def on_click_chat_info_wrapper(self, event):
        if self.on_click_chat_info is None:
            return

        pills = "".join(
            [
                f"""<div class='chat_document_pill'>{d['name']}</div>"""
                for d in self.current_chat["metadata"]["documents"]
            ]
        )

        grid_height = len(self.current_chat["metadata"]["documents"]) // 3

        markdown = "\n".join(
            [
                "To change configurations, start a new chat.\n",
                "**Uploaded Files**",
                f"<div class='pills_list'>{pills}</div><br />\n\n",
                "----",
                "**Source Storage**",
                f"""<span>{self.current_chat['metadata']['source_storage']}</span>\n""",
                "----",
                "**Assistant**",
                f"""<span>{self.current_chat['metadata']['assistant']}</span>\n""",
                "**Advanced configuration**",
                *[
                    f"- **{key.replace('_', ' ').title()}**: {value}"
                    for key, value in self.current_chat["metadata"]["params"].items()
                ],
            ]
        )

        self.on_click_chat_info(
            event,
            "Chat Info",
            [
                pn.pane.Markdown(
                    markdown,
                    dedent=True,
                    stylesheets=ui.stylesheets(
                        (":host", {"width": "100%"}),
                        (
                            ".pills_list",
                            {
                                # "background-color": "gold",
                                "display": "grid",
                                "grid-auto-flow": "row",
                                "row-gap": "10px",
                                "grid-template": f"repeat({grid_height}, 1fr) / repeat(3, 1fr)",
                                "max-height": "200px",
                                "overflow": "scroll",
                            },
                        ),
                        (
                            ".chat_document_pill",
                            {
                                "background-color": "rgb(241,241,241)",
                                "margin-left": "5px",
                                "margin-right": "5px",
                                "padding": "5px 15px",
                                "border-radius": "10px",
                                "color": "var(--accent-color)",
                                "width": "fit-content",
                                "grid-column": "span 1",
                            },
                        ),
                        ("ul", {"list-style-type": "none"}),
                    ),
                ),
            ],
        )

    def on_click_source_info_wrapper(self, event, sources):
        if self.on_click_chat_info is None:
            return

        source_infos = []
        for rank, source in enumerate(sources, 1):
            location = source["location"]
            if location:
                location = f": page(s) {location}"
            source_infos.append(
                (
                    f"<b>{rank}. {source['document']['name']}</b> {location}",
                    pn.pane.Markdown(source["content"], css_classes=["source-content"]),
                )
            )

        self.on_click_chat_info(
            event,
            "Source Info",
            [
                pn.pane.Markdown(
                    "This response was generated using the following data from the uploaded files: <br />",
                    dedent=True,
                    stylesheets=[""" hr { width: 94%; height:1px;  }  """],
                ),
                pn.layout.Accordion(
                    *source_infos,
                    header_background="transparent",
                    stylesheets=ui.stylesheets((":host", {"width": "100%"})),
                ),
            ],
        )

    def set_current_chat(self, chat):
        self.current_chat = chat

    def get_user_from_role(self, role: Literal["system", "user", "assistant"]) -> str:
        if role == "system":
            return "Ragna"
        elif role == "user":
            return cast(str, self.user)
        elif role == "assistant":
            return cast(str, self.current_chat["metadata"]["assistant"])
        else:
            raise RuntimeError

    async def chat_callback(
        self, content: str, user: str, instance: pn.chat.ChatInterface
    ):
        try:
            answer_stream = self.api_wrapper.answer(self.current_chat["id"], content)
            answer = await anext(answer_stream)

            message = RagnaChatMessage(
                answer["content"],
                role="assistant",
                user=self.get_user_from_role("assistant"),
                sources=answer["sources"],
                on_click_source_info_callback=self.on_click_source_info_wrapper,
            )
            yield message

            async for chunk in answer_stream:
                message.content_pane.object += chunk["content"]

        except Exception:
            yield RagnaChatMessage(
                (
                    "Sorry, something went wrong. "
                    "If this problem persists, please contact your administrator."
                ),
                role="system",
                user=self.get_user_from_role("system"),
            )

    @pn.depends("current_chat")
    def chat_interface(self):
        if self.current_chat is None:
            return

        return RagnaChatInterface(
            *[
                RagnaChatMessage(
                    message["content"],
                    role=message["role"],
                    user=self.get_user_from_role(message["role"]),
                    sources=message["sources"],
                    timestamp=message["timestamp"],
                    on_click_source_info_callback=self.on_click_source_info_wrapper,
                )
                for message in self.current_chat["messages"]
            ],
            callback=self.chat_callback,
            user=self.user,
            get_user_from_role=self.get_user_from_role,
            show_rerun=False,
            show_undo=False,
            show_clear=False,
            show_button_name=False,
            view_latest=True,
            sizing_mode="stretch_width",
            # TODO: Remove the parameter when
            #  https://github.com/holoviz/panel/issues/6115 is merged and released. We
            #  currently need it to avoid sending a message when the text input is
            #  de-focussed. But this also means we can't hit enter to send.
            auto_send_types=[],
            widgets=[
                pn.widgets.TextInput(
                    placeholder="Ask a question about the documents",
                    stylesheets=ui.stylesheets(
                        (
                            ":host input[type='text']",
                            {
                                "border": "none !important",
                                "box-shadow": "0px 0px 6px 0px rgba(0, 0, 0, 0.2)",
                                "padding": "10px 10px 10px 15px",
                            },
                        ),
                        (
                            ":host input[type='text']:focus",
                            {
                                "box-shadow": "0px 0px 8px 0px rgba(0, 0, 0, 0.3)",
                            },
                        ),
                    ),
                )
            ],
            card_params=dict(
                stylesheets=ui.stylesheets(
                    (":host", {"border": "none !important"}),
                    (
                        ".chat-feed-log",
                        {
                            "padding-right": "18%",
                            "margin-left": "18%",
                            "padding-top": "25px !important",
                        },
                    ),
                    (
                        ".chat-interface-input-container",
                        {
                            "margin-left": "19%",
                            "margin-right": "20%",
                            "margin-bottom": "20px",
                        },
                    ),
                )
            ),
            show_activity_dot=False,
        )

    @pn.depends("current_chat")
    def header(self):
        if self.current_chat is None:
            return

        current_chat_name = ""
        if self.current_chat is not None:
            current_chat_name = self.current_chat["metadata"]["name"]

        chat_name_header = pn.pane.HTML(
            f"<p>{current_chat_name}</p>",
            sizing_mode="stretch_width",
            stylesheets=[
                """ 

                        :host p {
                            max-width: 50%;
                            height:100%;

                            text-overflow: ellipsis;
                            white-space: nowrap;
                            overflow: hidden;
                            margin: 0px 0px 0px 10px;

                            font-size:20px;
                            text-decoration: underline;
                            text-underline-offset: 4px;

                            /* I don't understand why this is necessary to vertically align the text ... */
                            line-height:250%; 
                            
                        }
                        """
            ],
        )

        chat_documents_pills = []
        if (
            self.current_chat is not None
            and "metadata" in self.current_chat
            and "documents" in self.current_chat["metadata"]
        ):
            doc_names = [d["name"] for d in self.current_chat["metadata"]["documents"]]

            # FIXME: Instead of setting a hard limit of 20 documents here, this should
            #  scale automatically with the width of page
            #  See https://github.com/Quansight/ragna/issues/224
            for doc_name in doc_names[:20]:
                pill = pn.pane.HTML(
                    f"""<div class="chat_document_pill">{doc_name}</div>""",
                    stylesheets=[
                        """
                                                 :host {
                                                    background-color: rgb(241,241,241);
                                                    margin-top: 15px;
                                                    margin-left: 5px;   
                                                    margin-right: 5px;
                                                    padding: 5px 15px;
                                                    border-radius: 10px;
                                                    color:var(--accent-color);
                                                    
                                                 }   

                                                 """
                    ],
                )

                chat_documents_pills.append(pill)

        self.chat_info_button.name = f"{self.current_chat['metadata']['assistant']} | {self.current_chat['metadata']['source_storage']}"

        return pn.Row(
            chat_name_header,
            *chat_documents_pills,
            self.chat_info_button,
            stylesheets=[
                """:host {  
                                    background-color: #F9F9F9;
                                    border-bottom: 1px solid #EEEEEE;
                                    width: 100% !important;
                                    margin:0px;
                                    height:54px;
                                    overflow:hidden;
                                }

                                :host div {
                                    
                                    vertical-align: middle;
                                }
"""
            ],
        )

    def set_loading(self, is_loading):
        self.main_column.loading = is_loading

    def __panel__(self):
        self.main_column = pn.Column(
            self.header,
            self.chat_interface,
            sizing_mode="stretch_width",
            stylesheets=[
                """                    :host { 
                                            background-color: #F9F9F9;
                                            
                                            height:100%;
                                            max-width: 100%;
                                            margin-left: min(15px, 2%);
                                            border-left: 1px solid #EEEEEE;
                                            border-right: 1px solid #EEEEEE;
                                        }
                                """
            ],
        )

        return self.main_column



import abc
import os
import secrets
import time
from typing import cast

import jwt
import rich
from fastapi import HTTPException, Request, status
from fastapi.security.utils import get_authorization_scheme_param


class Authentication(abc.ABC):
    """Abstract base class for authentication used by the REST API."""

    @abc.abstractmethod
    async def create_token(self, request: Request) -> str:
        """Authenticate user and create an authorization token.

        Args:
            request: Request send to the `/token` endpoint of the REST API.

        Returns:
            Authorization token.
        """
        pass

    @abc.abstractmethod
    async def get_user(self, request: Request) -> str:
        """
        Args:
            request: Request send to any endpoint of the REST API that requires
                authorization.

        Returns:
            Authorized user.
        """
        pass


class RagnaDemoAuthentication(Authentication):
    """Demo OAuth2 password authentication without requirements.

    !!! danger

        As the name implies, this authentication is just for demo purposes and should
        not be used in production.
    """

    def __init__(self) -> None:
        msg = f"INFO:\t{type(self).__name__}: You can log in with any username"
        self._password = os.environ.get("RAGNA_DEMO_AUTHENTICATION_PASSWORD")
        if self._password is None:
            msg = f"{msg} and a matching password."
        else:
            msg = f"{msg} and the password {self._password}"
        rich.print(msg)

    _JWT_SECRET = os.environ.get(
        "RAGNA_DEMO_AUTHENTICATION_SECRET", secrets.token_urlsafe(32)[:32]
    )
    _JWT_ALGORITHM = "HS256"
    _JWT_TTL = int(os.environ.get("RAGNA_DEMO_AUTHENTICATION_TTL", 60 * 60 * 24 * 7))

    async def create_token(self, request: Request) -> str:
        """Authenticate user and create an authorization token.

        User name is arbitrary. Authentication is possible in two ways:

        1. If the `RAGNA_DEMO_AUTHENTICATION_PASSWORD` environment variable is set, the
           password is checked against that.
        2. Otherwise, the password has to match the user name.

        Args:
            request: Request send to the `/token` endpoint of the REST API. Must include
                the `"username"` and `"password"` as form data.

        Returns:
            Authorization [JWT](https://jwt.io/) that expires after one week.
        """
        async with request.form() as form:
            username = form.get("username")
            password = form.get("password")

        if username is None or password is None:
            raise HTTPException(status.HTTP_422_UNPROCESSABLE_ENTITY)

        if (self._password is not None and password != self._password) or (
            self._password is None and password != username
        ):
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

        return jwt.encode(
            payload={"user": username, "exp": time.time() + self._JWT_TTL},
            key=self._JWT_SECRET,
            algorithm=self._JWT_ALGORITHM,
        )

    async def get_user(self, request: Request) -> str:
        """Get user from an authorization token.

        Token has to be supplied in the
        [Bearer authentication scheme](https://swagger.io/docs/specification/authentication/bearer-authentication/),
        i.e. including a `Authorization: Bearer {token}` header.

        Args:
            request: Request send to any endpoint of the REST API that requires
                authorization.

        Returns:
            Authorized user.
        """

        unauthorized = HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not authenticated",
            headers={"WWW-Authenticate": "Bearer"},
        )

        authorization = request.headers.get("Authorization")
        scheme, token = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            raise unauthorized

        try:
            payload = jwt.decode(
                token, key=self._JWT_SECRET, algorithms=[self._JWT_ALGORITHM]
            )
        except (jwt.InvalidSignatureError, jwt.ExpiredSignatureError):
            raise unauthorized

        return cast(str, payload["user"])



import builtins
import sys
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Iterable,
    Iterator,
    Mapping,
    TypeVar,
    cast,
)

__all__ = [
    "itertools_pairwise",
    "importlib_metadata_package_distributions",
    "aiter",
    "anext",
]

T = TypeVar("T")


def _itertools_pairwise() -> Callable[[Iterable[T]], Iterator[tuple[T, T]]]:
    if sys.version_info[:2] >= (3, 10):
        from itertools import pairwise
    else:
        from itertools import tee
        from typing import Iterable, Iterator

        # https://docs.python.org/3/library/itertools.html#itertools.pairwise
        def pairwise(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
            # pairwise('ABCDEFG') --> AB BC CD DE EF FG
            a, b = tee(iterable)
            next(b, None)
            return zip(a, b)

    return pairwise


itertools_pairwise = _itertools_pairwise()


def _importlib_metadata_package_distributions() -> (
    Callable[[], Mapping[str, list[str]]]
):
    if sys.version_info[:2] >= (3, 10):
        from importlib.metadata import packages_distributions
    else:
        from importlib_metadata import packages_distributions

    return packages_distributions


importlib_metadata_package_distributions = _importlib_metadata_package_distributions()


def _aiter() -> Callable[[Any], AsyncIterator]:
    if sys.version_info[:2] >= (3, 10):
        aiter = builtins.aiter
    else:

        def aiter(obj: Any) -> AsyncIterator:
            return cast(AsyncIterator, obj.__aiter__())

    return aiter


aiter = _aiter()


def _anext() -> Callable[[AsyncIterator[T]], Awaitable[T]]:
    if sys.version_info[:2] >= (3, 10):
        anext = builtins.anext
    else:
        sentinel = object()

        def anext(
            ait: AsyncIterator[T],
            default: T = sentinel,  # type: ignore[assignment]
        ) -> Awaitable[T]:
            async def anext_impl() -> T:
                try:
                    return await ait.__anext__()
                except StopAsyncIteration:
                    if default is not sentinel:
                        return default

                    raise

            return anext_impl()

    return anext


anext = _anext()



import panel as pn
import param

from . import styles as ui

# TODO Move this into a CSS file
login_inputs_stylesheets = """

:host {
    width:100%;
    margin-left:0px;
    margin-right:0px;
}

label {
    font-weight: 600;
    font-size: 16px;
}

input {
    background-color: white !important;
}

"""


class AuthPage(pn.viewable.Viewer, param.Parameterized):
    feedback_message = param.String(default=None)

    custom_js = param.String(default="")

    def __init__(self, api_wrapper, **params):
        super().__init__(**params)
        self.api_wrapper = api_wrapper

        self.main_layout = None

        self.login_input = pn.widgets.TextInput(
            name="Email",
            stylesheets=[login_inputs_stylesheets, ui.BK_INPUT_GRAY_BORDER],
        )
        self.password_input = pn.widgets.PasswordInput(
            name="Password",
            stylesheets=[login_inputs_stylesheets, ui.BK_INPUT_GRAY_BORDER],
        )

    async def perform_login(self, event=None):
        self.main_layout.loading = True

        home_path = pn.state.location.pathname.rstrip("/").rstrip("auth")
        try:
            authed = await self.api_wrapper.auth(
                self.login_input.value, self.password_input.value
            )

            if authed:
                # Sets the cookie on the JS side
                self.custom_js = f""" document.cookie = "auth_token={self.api_wrapper.auth_token}; path:{home_path}";  """

        except Exception:
            authed = False

        if authed:
            # perform redirect
            pn.state.location.param.update(reload=True, pathname=home_path)
        else:
            self.feedback_message = "Authentication failed. Please retry."

        self.main_layout.loading = False

    @pn.depends("feedback_message")
    def display_error_message(self):
        if self.feedback_message is None:
            return None
        else:
            return pn.pane.HTML(
                f"""<div class="error">{self.feedback_message}</div>""",
                stylesheets=[
                    """
                                             :host { 
                                                width:100%;
                                                margin-left:0px;
                                                margin-right:0px;
                                             }  
                                             
                                             div.error { 
                                                width:100%; 
                                                color:red; 
                                                text-align:center;
                                                font-weight: 600;
                                                font-size: 16px;
                                             } """
                ],
            )

    @pn.depends("custom_js")
    def wrapped_custom_js(self):
        return pn.pane.HTML(
            f""" 
            <script>
                {self.custom_js}
            </script
            """,
        )

    def __panel__(self):
        login_button = pn.widgets.Button(
            name="Sign In",
            button_type="primary",
            stylesheets=[
                """ :host { 
                                                                width:100%;
                                                                margin-left:0px;
                                                                margin-right:0px;
                                                       } """
            ],
        )
        login_button.on_click(self.perform_login)

        self.main_layout = pn.Column(
            self.wrapped_custom_js,
            pn.pane.HTML(
                "<h1>Log In</h1>",
                stylesheets=[
                    """ :host { 
                                                width:100%;
                                                margin-left:0px;
                                                margin-right:0px;
                                                text-align:center;
                                            } 
                                         h1 { 
                                            font-weight: 600;
                                            font-size: 24px;
                                         }
                                         """
                ],
            ),
            self.display_error_message,
            self.login_input,
            self.password_input,
            pn.pane.HTML("<br />"),
            login_button,
            stylesheets=[
                """ :host { 
                                    background-color:white;
                                    /*background-color:gold;*/
                                    border-radius: 5px;
                                    box-shadow: lightgray 0px 0px 10px;
                                    padding: 0 25px 0 25px;
                             
                                    width:30%;
                                    min-width:360px;
                                    max-width:430px;
                                    
                                    margin-left: auto;
                                    margin-right: auto;
                                    margin-top: 10%;

                              }
                             :host > div {
                                margin-bottom: 10px;
                                margin-top: 10px;
                             }

                             .bk-panel-models-layout-Column {
                                width:100%;
                             }
                             
                             """
            ],
        )

        return self.main_layout



import os

import pytest

from ragna import assistants
from ragna._compat import anext
from ragna.assistants._api import ApiAssistant
from ragna.core import RagnaException
from tests.utils import skip_on_windows

API_ASSISTANTS = [
    assistant
    for assistant in assistants.__dict__.values()
    if isinstance(assistant, type)
    and issubclass(assistant, ApiAssistant)
    and assistant is not ApiAssistant
]


@skip_on_windows
@pytest.mark.parametrize("assistant", API_ASSISTANTS)
async def test_api_call_error_smoke(mocker, assistant):
    mocker.patch.dict(os.environ, {assistant._API_KEY_ENV_VAR: "SENTINEL"})

    chunks = assistant().answer(prompt="?", sources=[])

    with pytest.raises(RagnaException, match="API call failed"):
        await anext(chunks)



import json
from typing import AsyncIterator, cast

from ragna.core import PackageRequirement, Requirement, Source

from ._api import ApiAssistant


class OpenaiApiAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "OPENAI_API_KEY"
    _MODEL: str

    @classmethod
    def _extra_requirements(cls) -> list[Requirement]:
        return [PackageRequirement("httpx_sse")]

    @classmethod
    def display_name(cls) -> str:
        return f"OpenAI/{cls._MODEL}"

    def _make_system_content(self, sources: list[Source]) -> str:
        # See https://github.com/openai/openai-cookbook/blob/main/examples/How_to_format_inputs_to_ChatGPT_models.ipynb
        instruction = (
            "You are an helpful assistants that answers user questions given the context below. "
            "If you don't know the answer, just say so. Don't try to make up an answer. "
            "Only use the sources below to generate the answer."
        )
        return instruction + "\n\n".join(source.content for source in sources)

    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        import httpx_sse

        # See https://platform.openai.com/docs/api-reference/chat/create
        # and https://platform.openai.com/docs/api-reference/chat/streaming
        async with httpx_sse.aconnect_sse(
            self._client,
            "POST",
            "https://api.openai.com/v1/chat/completions",
            headers={
                "Content-Type": "application/json",
                "Authorization": f"Bearer {self._api_key}",
            },
            json={
                "messages": [
                    {
                        "role": "system",
                        "content": self._make_system_content(sources),
                    },
                    {
                        "role": "user",
                        "content": prompt,
                    },
                ],
                "model": self._MODEL,
                "temperature": 0.0,
                "max_tokens": max_new_tokens,
                "stream": True,
            },
        ) as event_source:
            await self._assert_api_call_is_success(event_source.response)

            async for sse in event_source.aiter_sse():
                data = json.loads(sse.data)
                choice = data["choices"][0]
                if choice["finish_reason"] is not None:
                    break

                yield cast(str, choice["delta"]["content"])


class Gpt35Turbo16k(OpenaiApiAssistant):
    """[OpenAI GPT-3.5](https://platform.openai.com/docs/models/gpt-3-5)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "gpt-3.5-turbo-16k"


class Gpt4(OpenaiApiAssistant):
    """[OpenAI GPT-4](https://platform.openai.com/docs/models/gpt-4)

    !!! info "Required environment variables"

        - `OPENAI_API_KEY`

    !!! info "Required packages"

        - `httpx_sse`
    """

    _MODEL = "gpt-4"



from typing import AsyncIterator, cast

from ragna.core import Source

from ._api import ApiAssistant


class Ai21LabsAssistant(ApiAssistant):
    _API_KEY_ENV_VAR = "AI21_API_KEY"
    _MODEL_TYPE: str

    @classmethod
    def display_name(cls) -> str:
        return f"AI21Labs/jurassic-2-{cls._MODEL_TYPE}"

    def _make_system_content(self, sources: list[Source]) -> str:
        instruction = (
            "You are a helpful assistant that answers user questions given the context below. "
            "If you don't know the answer, just say so. Don't try to make up an answer. "
            "Only use the sources below to generate the answer."
        )
        return instruction + "\n\n".join(source.content for source in sources)

    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        # See https://docs.ai21.com/reference/j2-chat-api#chat-api-parameters
        # See https://docs.ai21.com/reference/j2-complete-api-ref#api-parameters
        # See https://docs.ai21.com/reference/j2-chat-api#understanding-the-response
        response = await self._client.post(
            f"https://api.ai21.com/studio/v1/j2-{self._MODEL_TYPE}/chat",
            headers={
                "accept": "application/json",
                "content-type": "application/json",
                "authorization": f"Bearer {self._api_key}",
            },
            json={
                "numResults": 1,
                "temperature": 0.0,
                "maxTokens": max_new_tokens,
                "messages": [
                    {
                        "text": prompt,
                        "role": "user",
                    }
                ],
                "system": self._make_system_content(sources),
            },
        )
        await self._assert_api_call_is_success(response)

        yield cast(str, response.json()["outputs"][0]["text"])


# The Jurassic2Mid assistant receives a 500 internal service error from the remote
# server. See https://github.com/Quansight/ragna/pull/303
# TODO: Reinstate when the remote server is fixed
# class Jurassic2Mid(Ai21LabsAssistant):
#     """[AI21 Labs Jurassic-2 Mid](https://docs.ai21.com/docs/jurassic-2-models)
#
#     !!! info "Required environment variables"
#
#         - `AI21_API_KEY`
#     """
#
#     _MODEL_TYPE = "mid"


class Jurassic2Ultra(Ai21LabsAssistant):
    """[AI21 Labs Jurassic-2 Ultra](https://docs.ai21.com/docs/jurassic-2-models)

    !!! info "Required environment variables"

        - `AI21_API_KEY`
    """

    _MODEL_TYPE = "ultra"



import panel as pn


def preformat(text):
    """Allows {{key}} to be used for formatting in textcthat already uses
    curly braces.  First switch this into something else, replace curlies
    with double curlies, and then switch back to regular braces
    """
    text = text.replace("{{", "<<<").replace("}}", ">>>")
    text = text.replace("{", "{{").replace("}", "}}")
    text = text.replace("<<<", "{").replace(">>>", "}")
    return text


def redirect_script(remove, append="/", remove_auth_cookie=False):
    """
    This function returns a js script to redirect to correct url.
    :param remove: string to remove from the end of the url
    :param append: string to append at the end of the url
    :param remove_auth_cookie: boolean, will clear auth_token cookie when true.
    :return: string javascript script

    Examples:
    =========

    # This will remove nothing from the end of the url and will
    # add auth to it, so /foo/bar/car/ becomes /foo/bar/car/auth
    >>> redirect_script(remove="", append="auth")

    # This will remove nothing from the end of the url and will
    # add auth to it, so /foo/bar/car/ becomes /foo/bar/car/logout
    >>> redirect_script(remove="", append="logout")

    # This will remove "auth" from the end of the url and will add / to it
    # so /foo/bar/car/auth becomes /foo/bar/car/
    >>> redirect_script(remove="auth", append="/")
    """
    js_script = preformat(
        r"""
        <script>
            var currentPath = window.location.pathname; // Get the current path
            if (currentPath.includes('/{{remove}}')) {
              if ("{{remove}}") {
                currentPath = currentPath.replace(/\/{{remove}}(\/)?$/, '')
              }
              var redirectTo = currentPath + '{{append}}';
              window.location.href = redirectTo;
              if ({{remove_auth_cookie}}) {
                document.cookie = "auth_token=; expires=Thu, 01 Jan 1970 00:00:00 UTC; path=${redirectTo}";
              }
            }
        </script>
        """
    )

    return pn.pane.HTML(
        js_script.format(
            remove=remove,
            append=append,
            remove_auth_cookie=str(remove_auth_cookie).lower(),
        )
    )



from __future__ import annotations

import abc
import io
import os
import secrets
import time
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Optional, Type, TypeVar, Union

import jwt
from pydantic import BaseModel

from ._utils import PackageRequirement, RagnaException, Requirement, RequirementsMixin

if TYPE_CHECKING:
    from ragna.deploy import Config


class DocumentUploadParameters(BaseModel):
    method: str
    url: str
    data: dict


class Document(RequirementsMixin, abc.ABC):
    """Abstract base class for all documents."""

    def __init__(
        self,
        *,
        id: Optional[uuid.UUID] = None,
        name: str,
        metadata: dict[str, Any],
        handler: Optional[DocumentHandler] = None,
    ):
        self.id = id or uuid.uuid4()
        self.name = name
        self.metadata = metadata
        self.handler = handler or self.get_handler(name)

    @staticmethod
    def supported_suffixes() -> set[str]:
        """
        Returns:
            Suffixes, i.e. `".txt"`, that can be handled by the builtin
                [ragna.core.DocumentHandler][]s.
        """
        return set(DOCUMENT_HANDLERS.keys())

    @staticmethod
    def get_handler(name: str) -> DocumentHandler:
        """Get a document handler based on a suffix.

        Args:
            name: Name of the document.
        """
        handler = DOCUMENT_HANDLERS.get(Path(name).suffix)
        if handler is None:
            raise RagnaException

        return handler

    @classmethod
    @abc.abstractmethod
    async def get_upload_info(
        cls, *, config: Config, user: str, id: uuid.UUID, name: str
    ) -> tuple[dict[str, Any], DocumentUploadParameters]:
        pass

    @abc.abstractmethod
    def is_readable(self) -> bool:
        ...

    @abc.abstractmethod
    def read(self) -> bytes:
        ...

    def extract_pages(self) -> Iterator[Page]:
        yield from self.handler.extract_pages(self)


class LocalDocument(Document):
    """Document class for files on the local file system.

    !!! tip

        This object is usually not instantiated manually, but rather through
        [ragna.core.LocalDocument.from_path][].
    """

    @classmethod
    def from_path(
        cls,
        path: Union[str, Path],
        *,
        id: Optional[uuid.UUID] = None,
        metadata: Optional[dict[str, Any]] = None,
        handler: Optional[DocumentHandler] = None,
    ) -> LocalDocument:
        """Create a [ragna.core.LocalDocument][] from a path.

        Args:
            path: Local path to the file.
            id: ID of the document. If omitted, one is generated.
            metadata: Optional metadata of the document.
            handler: Document handler. If omitted, a builtin handler is selected based
                on the suffix of the `path`.

        Raises:
            RagnaException: If `metadata` is passed and contains a `"path"` key.
        """
        if metadata is None:
            metadata = {}
        elif "path" in metadata:
            raise RagnaException(
                "The metadata already includes a 'path' key. "
                "Did you mean to instantiate the class directly?"
            )

        path = Path(path).expanduser().resolve()
        metadata["path"] = str(path)

        return cls(id=id, name=path.name, metadata=metadata, handler=handler)

    @property
    def path(self) -> Path:
        return Path(self.metadata["path"])

    def is_readable(self) -> bool:
        return self.path.exists()

    def read(self) -> bytes:
        with open(self.path, "rb") as stream:
            return stream.read()

    _JWT_SECRET = os.environ.get(
        "RAGNA_API_DOCUMENT_UPLOAD_SECRET", secrets.token_urlsafe(32)[:32]
    )
    _JWT_ALGORITHM = "HS256"

    @classmethod
    async def get_upload_info(
        cls, *, config: Config, user: str, id: uuid.UUID, name: str
    ) -> tuple[dict[str, Any], DocumentUploadParameters]:
        url = f"{config.api.url}/document"
        data = {
            "token": jwt.encode(
                payload={
                    "user": user,
                    "id": str(id),
                    "exp": time.time() + 5 * 60,
                },
                key=cls._JWT_SECRET,
                algorithm=cls._JWT_ALGORITHM,
            )
        }
        metadata = {"path": str(config.local_root / "documents" / str(id))}
        return metadata, DocumentUploadParameters(method="PUT", url=url, data=data)

    @classmethod
    def decode_upload_token(cls, token: str) -> tuple[str, uuid.UUID]:
        try:
            payload = jwt.decode(
                token, key=cls._JWT_SECRET, algorithms=[cls._JWT_ALGORITHM]
            )
        except jwt.InvalidSignatureError:
            raise RagnaException(
                "Token invalid", http_status_code=401, http_detail=RagnaException.EVENT
            )
        except jwt.ExpiredSignatureError:
            raise RagnaException(
                "Token expired", http_status_code=401, http_detail=RagnaException.EVENT
            )
        return payload["user"], uuid.UUID(payload["id"])


class Page(BaseModel):
    """Dataclass for pages of a document

    Attributes:
        text: Text included in the page.
        number: Page number.
    """

    text: str
    number: Optional[int] = None


class DocumentHandler(RequirementsMixin, abc.ABC):
    """Base class for all document handlers."""

    @classmethod
    @abc.abstractmethod
    def supported_suffixes(cls) -> list[str]:
        """
        Returns:
            Suffixes supported by this document handler.
        """
        pass

    @abc.abstractmethod
    def extract_pages(self, document: Document) -> Iterator[Page]:
        """Extract pages from a document.

        Args:
            document: Document to extract pages from.

        Returns:
            Extracted pages.
        """
        ...


T = TypeVar("T", bound=DocumentHandler)


class DocumentHandlerRegistry(dict[str, DocumentHandler]):
    def load_if_available(self, cls: Type[T]) -> Type[T]:
        if cls.is_available():
            instance = cls()
            for suffix in cls.supported_suffixes():
                self[suffix] = instance

        return cls


DOCUMENT_HANDLERS = DocumentHandlerRegistry()


@DOCUMENT_HANDLERS.load_if_available
class PlainTextDocumentHandler(DocumentHandler):
    """Document handler for plain-text documents.
    Currently supports `.txt` and `.md` extensions.
    """

    @classmethod
    def supported_suffixes(cls) -> list[str]:
        return [".txt", ".md"]

    def extract_pages(self, document: Document) -> Iterator[Page]:
        yield Page(text=document.read().decode())


@DOCUMENT_HANDLERS.load_if_available
class PdfDocumentHandler(DocumentHandler):
    """Document handler for `.pdf` documents.

    !!! info "Package requirements"

        - [`pymupdf`](https://pymupdf.readthedocs.io/en/latest/)
    """

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [PackageRequirement("pymupdf>=1.23.6")]

    @classmethod
    def supported_suffixes(cls) -> list[str]:
        # TODO: pymudpdf supports a lot more formats, while .pdf is by far the most
        #  prominent. Should we expose the others here as well?
        return [".pdf"]

    def extract_pages(self, document: Document) -> Iterator[Page]:
        import fitz

        with fitz.Document(
            stream=document.read(), filetype=Path(document.name).suffix
        ) as document:
            for number, page in enumerate(document, 1):
                yield Page(text=page.get_text(sort=True), number=number)


@DOCUMENT_HANDLERS.load_if_available
class DocxDocumentHandler(DocumentHandler):
    """Document handler for `.docx` documents.

    !!! note

        This does *not* extract text from headers or footers.

    !!! info "Package requirements"

        - [`python-docx`](https://github.com/python-openxml/python-docx)
    """

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [PackageRequirement("python-docx")]

    @classmethod
    def supported_suffixes(cls) -> list[str]:
        return [".docx"]

    def extract_pages(self, document: Document) -> Iterator[Page]:
        import docx

        document_docx = docx.Document(io.BytesIO(document.read()))
        for paragraph in document_docx.paragraphs:
            text = paragraph.text
            if len(text) > 0:
                yield Page(text=text)


@DOCUMENT_HANDLERS.load_if_available
class PptxDocumentHandler(DocumentHandler):
    """Document handler for `.pptx` documents.

    !!! info "Package requirements"

        - [`python-pptx`](https://github.com/scanny/python-pptx)
    """

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [PackageRequirement("python-pptx")]

    @classmethod
    def supported_suffixes(cls) -> list[str]:
        return [".pptx"]

    def extract_pages(self, document: Document) -> Iterator[Page]:
        import pptx

        document_pptx = pptx.Presentation(io.BytesIO(document.read()))
        for number, slide in enumerate(document_pptx.slides, 1):
            text = "\n\n".join(
                shape.text
                for shape in slide.shapes
                if shape.has_text_frame and shape.text
            )
            yield Page(text=text, number=number)



from datetime import datetime, timedelta, timezone

import panel as pn
import param

from . import js
from . import styles as ui
from .components.file_uploader import FileUploader


def get_default_chat_name(timezone_offset=None):
    if timezone_offset is None:
        return f"Chat {datetime.now():%m/%d/%Y %I:%M %p}"
    else:
        tz = timezone(offset=timedelta(minutes=timezone_offset))
        return f"Chat {datetime.now().astimezone(tz=tz):%m/%d/%Y %I:%M %p}"


class ChatConfig(param.Parameterized):
    allowed_documents = param.List(default=["TXT"])

    source_storage_name = param.Selector()
    assistant_name = param.Selector()

    chunk_size = param.Integer(
        default=500,
        step=100,
        bounds=(100, 1_000),
    )
    chunk_overlap = param.Integer(
        default=250,
        step=50,
        bounds=(50, 500),
    )

    max_context_tokens = param.Integer(
        step=500,
        bounds=(1, 8000),
        default=2000,
        doc=(
            "Maximum number of context tokens and in turn the number of document chunks "
            "pulled out of the vector database."
        ),
    )

    max_new_tokens = param.Integer(
        default=1_000,
        step=100,
        bounds=(100, 10_000),
    )

    def is_assistant_disabled(self):
        return "Ragna/Demo" in self.assistant_name

    def is_source_storage_disabled(self):
        return "Ragna/Demo" in self.source_storage_name

    def to_params_dict(self):
        result = {}
        if not self.is_assistant_disabled():
            result["max_new_tokens"] = self.max_new_tokens

        if not self.is_source_storage_disabled():
            result["chunk_overlap"] = self.chunk_overlap
            result["chunk_size"] = self.chunk_size
            result["num_tokens"] = self.max_context_tokens

        return result


class ModalConfiguration(pn.viewable.Viewer):
    chat_name = param.String()

    config = param.ClassSelector(class_=ChatConfig, default=None)
    new_chat_ready_callback = param.Callable()
    cancel_button_callback = param.Callable()

    advanced_config_collapsed = param.Boolean(default=True)

    def __init__(self, api_wrapper, **params):
        super().__init__(chat_name=get_default_chat_name(), **params)

        self.api_wrapper = api_wrapper

        upload_endpoints = self.api_wrapper.upload_endpoints()

        self.chat_name_input = pn.widgets.TextInput.from_param(
            self.param.chat_name,
            stylesheets=[ui.BK_INPUT_GRAY_BORDER],
        )
        self.document_uploader = FileUploader(
            [],  # the allowed documents are set in the model_section function
            self.api_wrapper.auth_token,
            upload_endpoints["informations_endpoint"],
        )

        # Most widgets (including those that use from_param) should be placed after the super init call
        self.cancel_button = pn.widgets.Button(
            name="Cancel", button_type="default", min_width=375
        )
        self.cancel_button.on_click(self.cancel_button_callback)

        self.start_chat_button = pn.widgets.Button(
            name="Start Conversation", button_type="primary", min_width=375
        )
        self.start_chat_button.on_click(self.did_click_on_start_chat_button)

        self.upload_files_label = pn.pane.HTML()
        self.change_upload_files_label()

        self.upload_row = pn.Row(
            self.document_uploader,
            sizing_mode="stretch_width",
            stylesheets=[""" :host { margin-bottom: 20px; } """],
        )

        self.got_timezone = False

    def did_click_on_start_chat_button(self, event):
        if not self.document_uploader.can_proceed_to_upload():
            self.change_upload_files_label("missing_file")
        else:
            self.start_chat_button.disabled = True
            self.document_uploader.perform_upload(event, self.did_finish_upload)

    async def did_finish_upload(self, uploaded_documents):
        # at this point, the UI has uploaded the files to the API.
        # We can now start the chat

        try:
            new_chat_id = await self.api_wrapper.start_and_prepare(
                name=self.chat_name,
                documents=uploaded_documents,
                source_storage=self.config.source_storage_name,
                assistant=self.config.assistant_name,
                params=self.config.to_params_dict(),
            )

            self.start_chat_button.disabled = False

            if self.new_chat_ready_callback is not None:
                await self.new_chat_ready_callback(new_chat_id)

        except Exception:
            self.change_upload_files_label("upload_error")
            self.document_uploader.loading = False
            self.start_chat_button.disabled = False

    def change_upload_files_label(self, mode="normal"):
        if mode == "upload_error":
            self.upload_files_label.object = "<b>Upload files</b> (required)<span style='color:red;padding-left:100px;'><b>An error occured. Please try again or contact your administrator.</b></span>"
        elif mode == "missing_file":
            self.upload_files_label.object = (
                "<span style='color:red;'><b>Upload files</b> (required)</span>"
            )
        else:
            self.upload_files_label.object = "<b>Upload files</b> (required)"

    async def model_section(self):
        # prevents re-rendering the section
        if self.config is None:
            # Retrieve the components from the API and build a config object
            components = await self.api_wrapper.get_components()
            # TODO : use the components to set up the default values for the various params

            config = ChatConfig()
            config.allowed_documents = components["documents"]

            assistants = [component["title"] for component in components["assistants"]]

            config.param.assistant_name.objects = assistants
            config.assistant_name = assistants[0]

            source_storages = [
                component["title"] for component in components["source_storages"]
            ]
            config.param.source_storage_name.objects = source_storages
            config.source_storage_name = source_storages[0]

            # Now that the config object is set, we can assign it to the param.
            # This will trigger the update of the advanced_config_ui section
            self.config = config
            self.document_uploader.allowed_documents = config.allowed_documents

        return pn.Row(
            pn.Column(
                pn.pane.HTML("<b>Assistants</b>"),
                pn.widgets.Select.from_param(
                    self.config.param.assistant_name,
                    name="",
                    stylesheets=[ui.BK_INPUT_GRAY_BORDER],
                ),
            ),
            pn.Column(
                pn.pane.HTML("<b>Source storage</b>"),
                pn.widgets.Select.from_param(
                    self.config.param.source_storage_name,
                    name="",
                    stylesheets=[ui.BK_INPUT_GRAY_BORDER],
                ),
            ),
        )

    @pn.depends("config", "config.assistant_name", "config.source_storage_name")
    def advanced_config_ui(self):
        if self.config is None:
            return

        disabled_assistant = self.config.is_assistant_disabled()
        disabled_source_storage = self.config.is_source_storage_disabled()

        card = pn.Card(
            pn.Row(
                pn.Column(
                    pn.pane.HTML(
                        """<h2> Retrieval Method</h2>
                            <span>Adjusting these settings requires re-embedding the documents, which may take some time.
                            </span><br />
                                         """
                    ),
                    pn.widgets.IntSlider.from_param(
                        self.config.param.chunk_size,
                        name="Chunk Size",
                        bar_color=ui.MAIN_COLOR,
                        stylesheets=[ui.SS_LABEL_STYLE],
                        width_policy="max",
                        disabled=disabled_source_storage,
                        css_classes=["disabled"] if disabled_source_storage else [],
                    ),
                    pn.widgets.IntSlider.from_param(
                        self.config.param.chunk_overlap,
                        name="Chunk Overlap",
                        bar_color=ui.MAIN_COLOR,
                        stylesheets=[ui.SS_LABEL_STYLE],
                        width_policy="max",
                        disabled=disabled_source_storage,
                        css_classes=["disabled"] if disabled_source_storage else [],
                    ),
                    margin=(0, 20, 0, 0),
                    width_policy="max",
                ),
                pn.Column(
                    pn.pane.HTML(
                        """<h2> Model Configuration</h2>
                            <span>Changing these parameters alters the output. This might affect accuracy and efficiency.
                            </span><br />
                                         """
                    ),
                    pn.widgets.IntSlider.from_param(
                        self.config.param.max_context_tokens,
                        bar_color=ui.MAIN_COLOR,
                        stylesheets=[ui.SS_LABEL_STYLE],
                        width_policy="max",
                        disabled=disabled_source_storage,
                        css_classes=["disabled"] if disabled_source_storage else [],
                    ),
                    pn.widgets.IntSlider.from_param(
                        self.config.param.max_new_tokens,
                        bar_color=ui.MAIN_COLOR,
                        stylesheets=[ui.SS_LABEL_STYLE],
                        width_policy="max",
                        disabled=disabled_assistant,
                        css_classes=["disabled"] if disabled_assistant else [],
                    ),
                    width_policy="max",
                    height_policy="max",
                    margin=(0, 20, 0, 0),
                    styles={
                        "border-left": "1px solid var(--neutral-stroke-divider-rest)"
                    },
                ),
                height=250,
            ),
            collapsed=self.advanced_config_collapsed,
            collapsible=True,
            hide_header=True,
            stylesheets=[ui.SS_ADVANCED_UI_CARD],
        )

        def toggle_card(event):
            if event.old < event.new:
                # This callback is triggered when the card is rerendered,
                # after changing the assistant, for example.
                # This test prevents collapsing the card when it is not needed

                card.collapsed = not card.collapsed
                self.advanced_config_collapsed = card.collapsed

            toggle_button = event.obj

            if card.collapsed:
                toggle_button.name = toggle_button.name.replace("▼", "▶")
            else:
                toggle_button.name = toggle_button.name.replace("▶", "▼")

        toggle_button = pn.widgets.Button(
            name="Advanced Configurations   ▶",
            button_type="light",
            stylesheets=[
                """button.bk-btn { 
                        font-size:13px; 
                        font-weight:600; 
                        padding-left: 0px;
                        color: MAIN_COLOR; 
                }""".replace("MAIN_COLOR", ui.MAIN_COLOR)
            ],
        )

        toggle_button.on_click(toggle_card)

        toggle_button.js_on_click(
            args={"card": card},
            code=js.JS_TOGGLE_CARD,
        )

        return pn.Column(toggle_button, card)

    def __panel__(self):
        return pn.Column(
            pn.pane.HTML(
                f"""<h2>Start a new chat</h2>
                         Let's set up the configurations for your new chat !<br />
                         <script>{js.reset_modal_size(ui.CONFIG_MODAL_WIDTH, ui.CONFIG_MODAL_MIN_HEIGHT)}</script>
                         """,
            ),
            ui.divider(),
            pn.pane.HTML("<b>Chat name</b>"),
            self.chat_name_input,
            ui.divider(),
            self.model_section,
            ui.divider(),
            self.advanced_config_ui,
            ui.divider(),
            self.upload_files_label,
            self.upload_row,
            pn.Row(self.cancel_button, self.start_chat_button),
            min_height=ui.CONFIG_MODAL_MIN_HEIGHT,
            min_width=ui.CONFIG_MODAL_WIDTH,
            sizing_mode="stretch_both",
            height_policy="max",
        )
